/* This software is licensed by the MIT License, see LICENSE file */
/*                                Copyright © 2022 Gregory Lirent */

#include "../../include/set.h"
#include "../__internal/rbtree.h"

/*#####################################################################################################################*/

static inline int rbnode_compare(const rbnode_t* s0, const rbnode_t* s1, vtype t) {
    return vnode_compare(s0->value, t, s1->value, t);
}

static inline hash_t rbnode_hash(const rbnode_t* s, vtype t) {
    return vnode_hash(&s->value, t);
}

/*#####################################################################################################################*/

hash_t vset_hash(const set_t* s) {

    stack_t         z;
    rbnode_t *c0, *c1;
    hash_t    hash, v;

    if (rbnode_is_empty(s->root))
        return 0;

    stack_init(&z);
    stack_push(&z, s->root->left);

    hash = 1;

    if (!rbnode_is_empty(c0 = stack_pop(&z))) {
        do {
            ++hash;
            if (!rbnode_is_empty(c0->right)) stack_push(&z,      c0->right);
            if (!rbnode_is_empty(c0->left))  stack_push(&z, c1 = c0->left);
        } while (!is_null(c0 = stack_pop(&z)));
    }

    v  = rbnode_hash(c1, s->type);
    stack_push(&z, s->root->right);

    if (!rbnode_is_empty(c0 = stack_pop(&z))) {
        do {
            ++hash;
            if (!rbnode_is_empty(c0->right)) stack_push(&z, c1 = c0->right);
            if (!rbnode_is_empty(c0->left))  stack_push(&z,      c0->left);
        } while (!is_null(c0 = stack_pop(&z)));
    }

    v += rbnode_hash(c1, s->type);

    return (hash ^ v) + VTYPE_SET;
}

/*#####################################################################################################################*/

void vset_init(set_t* x, vtype t) {
    x->root = rbnode_empty;
    x->type = t;
}


void vset_free(set_t* x) {
    rbnode_t *t, *c;

    c = x->root;

    while (!rbnode_is_empty(x->root)) {
        if (!rbnode_is_empty(c->left)) {
            c = c->left;
        } else if (!rbnode_is_empty(c->right)) {
            c = c->right;
        } else if (!rbnode_is_root(c)) {
            vnode_free(&c->value, x->type);

            t = c;
            c = c->parent;

            if (t == c->left) c->left  = rbnode_empty;
            else              c->right = rbnode_empty;

            free(t);
        } else {
            vnode_free(&c->value, x->type);
            x->root = rbnode_empty;
        }
    }

    x->type = 0;
}

/*#####################################################################################################################*/

size_t vset_size(const set_t* x) {
    stack_t   z;
    size_t    n;
    rbnode_t* c;

    stack_init(&z);
    stack_push(&z, x->root);

    n = 0;

    if (!rbnode_is_empty(x->root)) {
        while ((c = stack_pop(&z))) {
            ++n;
            if (!rbnode_is_empty(c->right)) stack_push(&z, c->right);
            if (!rbnode_is_empty(c->left))  stack_push(&z, c->left);
        }
    }

    return n;
}

/*#####################################################################################################################*/

int vset_compare(const set_t* s0, const set_t* s1) {
    stack_t         z;
    rbnode_t *c0, *c1;
    int           cmp;

    if (s0 == s1 || s0->root == s1->root)
        return 0;
    if (s0->type != s1->type)
        return s0->type - s1->type;

    stack_init(&z);
    stack_push_many(&z, 2, (void*)s1, (void*)s0);

    cmp = 0;

    do {
        c0 = stack_pop(&z);
        c1 = stack_pop(&z);

        if (rbnode_is_empty(c0) || rbnode_is_empty(c1)) {
            if (c0 != c1) {
                stack_flush(&z);
                return rbnode_is_empty(c0) ? -1 : 1;
            }
        } else if ((cmp = rbnode_compare(c0, c1, s0->type))) {
            if (c0->left == c1->right) {  // == rbnode_empty
                          cmp = rbnode_compare(c0->right, c1, s0->type);
                if (!cmp) cmp = rbnode_compare(c0, c1->left, s0->type);
            } else if (c0->right == c1->left) { // == rbnode_empty
                          cmp = rbnode_compare(c0, c1->right, s0->type);
                if (!cmp) cmp = rbnode_compare(c0->left, c1, s0->type);
            }

            if (cmp) {
                stack_flush(&z);
                return cmp;
            }
        } else stack_push_many(&z, 4, c1->right, c0->right, c1->left, c0->left);

    } while (!is_null(z.value));

    return 0;
}
